package GD.Data;

import java.awt.Point;

/**
 * Coordinate factory. It manages coordinates conversion between tracker, client
 * and server coordinate systems. It is implemented as a singleton and requires
 * to be initialized by {@link JGDVncCanvas JGDVncCanvas} instance.
 *
 * @version 2.0.0
 * @author Andrey Shipilov
 *
 */
public class Converter {

    /**
     * Instance of the coordinate converter, it contains information, needed for
     * scaling and calculating offset of the coordinates
     */
    private static ICoordinateConverter mCnv = null;
    
    public static boolean IsInsideDisplay(int x, int y){
        return  x >= 0 && 
                y >= 0 &&
                x < mCnv.Width() && 
                y < mCnv.Height();
    }
    
    public static boolean IsInsideDisplay(Point pnt){
        return  IsInsideDisplay(pnt.x, pnt.y);
    }

    /**
     * Convert client coordinates to the display coordinates. Display coordinate
     * system origin is left upper corner of the screen (0, 0). There is no
     * difference in scaling between display and client coordinates. Only offset
     * should be applied.
     *
     * @param x X coordinate
     * @param y Y coordinate
     * @return Point in display (tracker) coordinate system
     */
    public static Point ClientToDisplay(int x, int y) {
        if (isInitialized()) {
            x = x + mCnv.XOffset();
            y = y + mCnv.YOffset();
        }

        return new Point(x, y);
    }

    /**
     * Convert client coordinates to the display coordinates. Display coordinate
     * system origin is left upper corner of the screen (0, 0). There is no
     * difference in scaling between display and client coordinates. Only offset
     * should be applied.
     *
     * @param pnt Point to be converted
     * @return Point in display (tracker) coordinate system
     */
    public static Point ClientToDisplay(Point pnt) {
        return ClientToDisplay(pnt.x, pnt.y);
    }

    public static Point ClientToDisplay(ICoordinates coord) {
        return ClientToDisplay(coord.Coords());
    }

    /**
     * Convert client coordinates to the server coordinates. Client coordinates
     * have the same origin as server. Only scaling should be applied.
     *
     * @param x X coordinate
     * @param y Y coordinate
     * @return Point in server coordinate system
     */
    public static Point ClientToServer(int x, int y) {
        if (isInitialized()) {
            x = (int) (x * 100 / mCnv.XScale());
            y = (int) (y * 100 / mCnv.YScale());
        }

        return new Point(x, y);
    }

    /**
     * Convert client coordinates to the server coordinates. Client coordinates
     * have the same origin as server. Only scaling should be applied.
     *
     * @param pnt Point to be converted
     * @return Point in server coordinate system
     */
    public static Point ClientToServer(Point pnt) {
        return ClientToServer(pnt.x, pnt.y);
    }

    public static Point ClientToServer(ICoordinates coord) {
        return ClientToServer(coord.Coords());
    }

    /**
     * Convert display (tracker) coordinates to the client coordinates. Display
     * coordinate system origin is left upper corner of the screen (0, 0). There
     * is no difference in scaling between display and client coordinates. Only
     * offset should be calculated.
     *
     * @param x X coordinate
     * @param y Y coordinate
     * @return Point in client coordinate system
     */
    public static Point DisplayToClient(int x, int y) {
        if (isInitialized()) {
            x = x - mCnv.XOffset();
            y = y - mCnv.YOffset();
        }

        return new Point(x, y);
    }

    /**
     * Convert display (tracker) coordinates to the client coordinates. Display
     * coordinate system origin is left upper corner of the screen (0, 0). There
     * is no difference in scaling between display and client coordinates. Only
     * offset should be calculated.
     *
     * @param pnt Point to be converted
     * @return Point in client coordinate system
     */
    public static Point DisplayToClient(Point pnt) {
        return DisplayToClient(pnt.x, pnt.y);
    }

    public static Point DisplayToClient(ICoordinates coord) {
        return DisplayToClient(coord.Coords());
    }

    /**
     * Convert display coordinates to the server coordinates. It converts
     * display -> client -> server coordinates.
     *
     * @param x X coordinate
     * @param y Y coordinate
     * @return Point in server coordinate system
     */
    public static Point DisplayToServer(int x, int y) {
        return ClientToServer(DisplayToClient(x, y));
    }

    /**
     * Convert display coordinates to the server coordinates. It converts
     * display -> client -> server coordinates.
     *
     * @param pnt Point to be converted
     * @return Point in server coordinate system
     */
    public static Point DisplayToServer(Point pnt) {
        return DisplayToServer(pnt.x, pnt.y);
    }

    public static Point DisplayToServer(ICoordinates coord) {
        return DisplayToServer(coord.Coords());
    }

    /**
     * Initialize factory with canvas instance. No conversions will be performed
     * before initialization
     *
     * @param cnv
     *            {@link JGDVncCanvas JGDVncCanvas} instance
     */
    public static void initialize(ICoordinateConverter cnv) {
        if (!isInitialized()) {
            mCnv = cnv;
        }
    }

    /**
     * Check if the factory was already initialized
     *
     * @return false if canvas instance was not initialized yet
     */
    public static boolean isInitialized() {
        return mCnv != null;
    }

    /**
     * Convert server coordinates to the client coordinates. This coordinates
     * are used to draw gaze mark upon shared desktop image. No offset should be
     * applied as origin of coordinate system of shared area is the same as the
     * origin of canvas.
     *
     * @param x X coordinate
     * @param y Y coordinate
     * @return Point in client coordinate system
     */
    public static Point ServerToClient(int x, int y) {
        if (isInitialized()) {
            x = (int) (x * mCnv.XScale() / 100);
            y = (int) (y * mCnv.YScale() / 100);
        }

        return new Point(x, y);
    }

    /**
     * Convert server coordinates to the client coordinates. This coordinates
     * are used to draw gaze mark upon shared desktop image. No offset should be
     * applied as origin of coordinate system of shared area is the same as the
     * origin of canvas.
     *
     * @param pnt Point to be converted
     * @return Point in client coordinate system
     */
    public static Point ServerToClient(Point pnt) {
        return ServerToClient(pnt.x, pnt.y);
    }

    public static Point ServerToClient(ICoordinates coord) {
        return ServerToClient(coord.Coords());
    }

    /**
     * Convert server coordinates to the display coordinates. It converts server
     * -> client -> display coordinates.
     *
     * @param x X coordinate
     * @param y Y coordinate
     * @return Point in display (tracker) coordinate system
     */
    public static Point ServerToDisplay(int x, int y) {
        return ClientToDisplay(ServerToClient(x, y));
    }

    /**
     * Convert server coordinates to the display coordinates. It converts server
     * -> client -> display coordinates.
     *
     * @param pnt Point to be converted
     * @return Point in display (tracker) coordinate system
     */
    public static Point ServerToDisplay(Point pnt) {
        return ServerToDisplay(pnt.x, pnt.y);
    }

    public static Point ServerToDisplay(ICoordinates coord) {
        return ServerToDisplay(coord.Coords());
    }

    public static int StringToInt(String txt, int defValue) {
        int ret;
        try {
            ret = Integer.parseInt(txt);
        } catch (Exception ex) {
            ret = defValue;
        }

        return ret;
    }
}
