package GazeVnc;

import GD.Data.*;
import java.awt.Point;
import java.io.IOException;
import java.util.Vector;

import TightVnc.RfbProto;
import TightVnc.VncViewer;

public class GazeProto extends RfbProto {

    /**
     * Message destination Id. Server is a destination, not for further distribution
     */
    final static int SERVER_ID = 65000;
    /**
     * Message destination Id. All clients are marked as destination
     */
    final static int ANY_ID = 64000;
    
    
    /**
     * Message types
     */
    final public static int gdGazeMsg = 201, //send gaze message
            gdControlMsg = 202, //send control message
            gdTextMsg = 203;	//send text
    /**
     * gdGazeMsg Message subtypes
     */
    final public static int gdGazeSample = 1, //send gaze message
            gdGazeFixation = 2, //send control message
            gdGazeBlink = 3;	//send text
    /**
     * gdControlMsg Message subtypes
     */
    final public static int gdPoke = 1, //poke client
            gdRegister = 2, //register client
            gdUnregister = 3;	//unregister client
    /**
     * gdTextMsg Message subtypes
     */
    final public static int gdName = 1, //name
            gdDescription = 2, //description
            gdText = 3;	//text message

    /**
     * Constructor
     *
     * @param h Host address
     * @param p Protocol
     * @param v Viewer
     * @throws IOException
     */
    GazeProto(String h, int p, GazeVncViewer v) throws IOException {
        super(h, p, v);
        
        
    }
    /**
     * S2C messages processing. First byte of the message already read to
     * identify its type.
     */
    /**
     * Last received image source
     */
    public short srcClientId;

    public int readServerMessageSubtype() throws IOException {
        return readU8();
    }

    /**
     * Read income server message, heading a list of gaze data structures type
     * and subtype already read
     *
     * @return
     * @throws IOException
     */
    private int readHeaderMsg() throws IOException {
        srcClientId = (short) readU16();
        skipBytes(2); //skip pad2

        return readU16(); //return number of gaze data structs
    }

    /**
     * Read one sample structure
     *
     * @return {@link JGDSingleSample JGDSingleSample} object
     * @throws IOException
     */
    private GazeEntrySample readSample() throws IOException {
        int ordinal = Integer.reverseBytes(readU32());
        int timestamp = Integer.reverseBytes(readU32());
        int eye = Short.reverseBytes((short) readU16());
        skipBytes(2);
        int x = Integer.reverseBytes(readU32());
        int y = Integer.reverseBytes(readU32());
        int ps = Integer.reverseBytes(readU32());

        return new GazeEntrySample(ordinal, timestamp, 0, eye, ps, new Point(x, y));
    }

    /**
     * Read one fixation structure
     *
     * @return {@link JGDFixation JGDFixation} object
     * @throws IOException
     */
    private GazeEntryFixation readFixation() throws IOException {
        int ordinal = Integer.reverseBytes(readU32());
        int time = Integer.reverseBytes(readU32());
        int duration = Integer.reverseBytes(readU32());
        int eye = Short.reverseBytes((short) readU16());
        readU16();
        int x = Integer.reverseBytes(readU32());
        int y = Integer.reverseBytes(readU32());

        return new GazeEntryFixation(ordinal, time, duration, eye, new Point(x, y));
    }

    /**
     * Read one blink structure
     *
     * @return {@link JGDBlink JGDBlink} object
     * @throws IOException
     */
    private GazeEntryBlink readBlink() throws IOException {
        int ordinal = Integer.reverseBytes(readU32());
        int time = Integer.reverseBytes(readU32());
        int duration = Integer.reverseBytes(readU32());
        int eye = Short.reverseBytes((short) readU16());
        skipBytes(2);

        return new GazeEntryBlink(ordinal, time, duration, eye);
    }

    /**
     * Read one blink structure
     *
     * @return {@link JGDBlink JGDBlink} object
     * @throws IOException
     */
    private String readText(int length) throws IOException {
        byte[] text = new byte[length];
        readFully(text);
        String str = new String(text);
        return str;
    }

    /**
     * Read single samples message
     *
     * @return Vector of {@link JGDSingleSample JGDSingleSample} structures,
     * received from the server
     * @throws IOException
     */
    public Vector<GazeEntrySample> ReadSamplesMsg() throws IOException {
        int cnt = readHeaderMsg();

        Vector<GazeEntrySample> ssVec = new Vector<GazeEntrySample>();
        while (cnt-- > 0) {
            ssVec.add(readSample());
        }

        return ssVec;
    }

    /**
     * Read fixations message
     *
     * @return Vector of {@link JGDFixation JGDFixation} structures, received
     * from the server
     * @throws IOException
     */
    public Vector<GazeEntryFixation> ReadFixationsMsg() throws IOException {
        int cnt = readHeaderMsg();

        Vector<GazeEntryFixation> fVec = new Vector<GazeEntryFixation>();
        while (cnt-- > 0) {
            fVec.add(readFixation());
        }

        return fVec;
    }

    /**
     * Read blinks message
     *
     * @return Vector of {@link JGDBlink JGDBlink} structures, received from the
     * server
     * @throws IOException
     */
    public Vector<GazeEntryBlink> ReadBlinksMsg() throws IOException {
        int cnt = readHeaderMsg();

        Vector<GazeEntryBlink> bVec = new Vector<GazeEntryBlink>();
        while (cnt-- > 0) {
            bVec.add(readBlink());
        }

        return bVec;
    }
    
    /**
     * Read register message @throws IOException
     */
    public short ReadRegisterMsg() throws IOException {
        readHeaderMsg();
        return srcClientId;
    }

    /**
     * Read unregister message @throws IOException
     */
    public short ReadUnRegisterMsg() throws IOException {
        readHeaderMsg();
        return srcClientId;
    }

    /**
     * Read register message @throws IOException
     */
    public short ReadPokeMsg() throws IOException {
        readHeaderMsg();
        return srcClientId;
    }
    
    /**
     * Read name message @throws IOException
     */
    public String ReadNameMsg() throws IOException {
        String name = readText(readHeaderMsg());
        return name;
    }
    
    /**
     * Read description message @throws IOException
     */
    public String ReadDescriptionMsg() throws IOException {
        return readText(readHeaderMsg());
    }
    
    /**
     * Read text message @throws IOException
     */
    public String ReadTextMsg() throws IOException {
        return readText(readHeaderMsg());
    }

    /**
     * C2S messages processing. Writing messages to the stream
     */
    private void writeHeaderMsg(int type, int subtype, int destinationId, int length) throws IOException {
        byte[] b = new byte[8];

        b[0] = (byte) type;
        b[1] = (byte) subtype;
        b[2] = (byte) (destinationId & 0xff);
        b[3] = (byte) ((destinationId >> 8) & 0xff);
        b[4] = (byte) 0;
        b[5] = (byte) 0;
        b[6] = (byte) (length & 0xff);
        b[7] = (byte) ((length >> 8) & 0xff);

        writeFully(b);
    }

    /**
     * Write outcome message, describing the data that will follow it
     *
     * @return
     * @throws IOException
     */
    private void WriteClientHeaderMsg(int msgType, int destinationId, int length) throws IOException {
        byte[] b = new byte[8];

        b[0] = (byte) msgType;
        b[1] = (byte) 0;
        b[2] = (byte) (destinationId & 0xff);
        b[3] = (byte) ((destinationId >> 8) & 0xff);
        b[4] = (byte) 0;
        b[5] = (byte) 0;
        b[6] = (byte) (length & 0xff);
        b[7] = (byte) ((length >> 8) & 0xff);

        writeFully(b);
    }

    private void writeGazeEntry(GazeEntryBase entry) throws IOException {
        writeFully(entry.MessageStruct());
    }

    /**
     * Write text
     *
     * @return {@link JGDBlink JGDBlink} object
     * @throws IOException
     */
    private void writeText(String str) throws IOException {
        writeFully(str.getBytes());
    }

    /**
     * Write single samples message
     *
     * @param ssVec Queue of {@link JGDSingleSample JGDSingleSample} structures,
     * received from the server
     * @throws IOException
     */
    public void WriteSamplesMsg(Queue<GazeEntryBase> sQue) throws IOException {
        synchronized (inlock) {
            if (!sQue.isEmpty()) {
                writeHeaderMsg(gdGazeMsg, gdGazeSample, ANY_ID, sQue.size());

                while (!sQue.isEmpty()) {
                    writeFully(sQue.extract().MessageStruct());
                }
            }
        }
    }

    synchronized protected void WriteGazeEntryMsg(Queue<GazeEntryBase> entries, int msgType, int entrySize, int dstId) throws IOException {
        int size = entries.size();
        if (size > 0) {
            byte[] b = new byte[8 + entrySize * size];

            b[0] = (byte) msgType;
            b[1] = (byte) 0;
            b[2] = (byte) (dstId & 0xff);
            b[3] = (byte) ((dstId >> 8) & 0xff);
            b[4] = (byte) 0;
            b[5] = (byte) 0;
            b[6] = (byte) (size & 0xff);
            b[7] = (byte) ((size >> 8) & 0xff);

            int pos = 8;
            while (!entries.isEmpty()) {
                System.arraycopy(entries.extract().MessageStruct(), 0, b, pos, entrySize);
                pos += entrySize;
            }

            writeFully(b);
        }
    }

    /**
     * Write fixations message
     *
     * @param fVec Queue of {@link JGDFixation JGDFixation} structures, received
     * from the server
     * @throws IOException
     */
    public void WriteFixationsMsg(Queue<GazeEntryBase> fQue) throws IOException {
        synchronized (inlock) {
            if (!fQue.isEmpty()) {
                writeHeaderMsg(gdGazeMsg, gdGazeFixation, ANY_ID, fQue.size());

                while (!fQue.isEmpty()) {
                    writeFully(fQue.extract().MessageStruct());
                }
            }
        }
    }

    /**
     * Write blinks message
     *
     * @param bVec Queue of {@link JGDBlink JGDBlink} structures, received from
     * the server
     * @throws IOException
     */
    public void WriteBlinksMsg(Queue<GazeEntryBase> bQue) throws IOException {
        synchronized (inlock) {
            if (!bQue.isEmpty()) {
                writeHeaderMsg(gdGazeMsg, gdGazeBlink, ANY_ID, bQue.size());

                while (!bQue.isEmpty()) {
                    writeFully(bQue.extract().MessageStruct());
                }
            }
        }
    }

    /**
     * Write name message
     *
     * @param name New name of the client that should be sent to the server and
     * distributed to clients
     * @throws IOException
     */
    public void WriteNameMsg(String name) throws IOException {
        synchronized (inlock) {
            writeHeaderMsg(gdTextMsg, gdName, ANY_ID, name.length());
            writeText(name);
        }
    }

    /**
     * Write description message
     *
     * @param description New description of the client that should be sent to
     * the server and distributed to clients
     * @throws IOException
     */
    public void WriteDescriptionMsg(String description) throws IOException {
        synchronized (inlock) {
            writeHeaderMsg(gdTextMsg, gdDescription, ANY_ID, description.length());
            writeText(description);
        }
    }

    /**
     * Write text message
     *
     * @param text New text message that should be sent to the server and
     * distributed to clients
     * @throws IOException
     */
    public void WriteTextMsg(String text) throws IOException {
        synchronized (inlock) {
            writeHeaderMsg(gdTextMsg, gdText, ANY_ID, text.length());
            writeText(text);
        }
    }

    /**
     * Write register message
     *
     * @throws IOException
     */
    public void WriteRegisterMsg() throws IOException {
        synchronized (inlock) {
            writeHeaderMsg(gdControlMsg, gdRegister, ANY_ID, 0);
        }
    }

    /**
     * Write unregister message
     *
     * @throws IOException
     */
    public void WriteUnregisterMsg() throws IOException {
        synchronized (inlock) {
            writeHeaderMsg(gdControlMsg, gdUnregister, ANY_ID, 0);
        }
    }

    /**
     * Write poke message
     *
     * @throws IOException
     */
    public void WritePokeMsg() throws IOException {
        synchronized (inlock) {
            writeHeaderMsg(gdControlMsg, gdPoke, ANY_ID, 0);
        }
    }
}
